// SimpleViewerDoc.cpp : implementation of the CSimpleViewerDoc class
//

#include "stdafx.h"
#include "SimpleViewer.h"
#include "SimpleViewerView.h"
#include "SimpleViewerDoc.h"

// SceniX ---------------------- begin
#include "nvsg/Node.h"
#include "nvsg/Triangles.h"
#include "nvsg/Transform.h"
#include "nvsg/GeoNode.h"
#include "nvsg/PointLight.h"
#include "nvsg/Material.h"
#include "nvsg/FaceAttribute.h"
#include "nvsg/StateSet.h"

//#define LOD_TEST
//#define LOD_EMPTYGROUPTEST
#ifdef LOD_TEST
#include "nvsg/LOD.h"
#endif

using namespace nvmath;
using namespace nvsg;
using namespace nvutil;

// enable memory leak detection
#include "nvutil/DbgNew.h"

// SceniX ---------------------- end


// CSimpleViewerDoc

IMPLEMENT_DYNCREATE(CSimpleViewerDoc, CDocument)

BEGIN_MESSAGE_MAP(CSimpleViewerDoc, CDocument)
END_MESSAGE_MAP()


// CSimpleViewerDoc construction/destruction

CSimpleViewerDoc::CSimpleViewerDoc()
: m_sceneHdl(NULL)
{
	// TODO: add one-time construction code here

}

CSimpleViewerDoc::~CSimpleViewerDoc()
{
}

BOOL CSimpleViewerDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
  {
    return FALSE;
  }

  // SceniX ---------------------- begin
 
  // Create a new scene and increment the ref count, 
  // because you want to use it.
  m_sceneHdl = CreateHandle(SceneHandle);
  m_sceneHdl->addRef();
  
  // Tell the scene which tree to use by providing the 
  // root node of the scene. CreateSimpleScene returns 
  // the root of the tree.
  // Note: 
  // setRootNode() increments the refcount of the 
  // given root node.
  WritableObject<Scene>(m_sceneHdl)->setRootNode( createSimpleTree() );

  // SceniX ---------------------- end

  return TRUE;
}

// SceniX ---------------------- begin
// This method creates a more or less complex tree and 
// returns a pointer to the root node of the tree.
NodeHandle * CSimpleViewerDoc::createSimpleTree() const
{
#ifndef LOD_TEST
  // Create the root node and apply a rotation.
  // Note: 
  // We do not increment the ref count here, 
  // because we throw the root into the scene 
  // via setRoot(), and setRoot() will increment 
  // it for us.
  // setup vertices and faces  
  Vec3f vertices[8] = { Vec3f(-0.2f, -0.2f,  0.2f)
    , Vec3f(-0.2f,  0.2f,  0.2f)
    , Vec3f( 0.2f,  0.2f,  0.2f)
    , Vec3f( 0.2f, -0.2f,  0.2f)
    , Vec3f( 0.2f, -0.2f, -0.2f)
    , Vec3f( 0.2f,  0.2f, -0.2f)
    , Vec3f(-0.2f,  0.2f, -0.2f)
    , Vec3f(-0.2f, -0.2f, -0.2f) };

  Face3 faces[12]     = { {0,2,1}, {2,0,3}
  , {3,5,2}, {5,3,4}
  , {4,6,5}, {6,4,7}
  , {7,1,6}, {1,7,0}
  , {0,4,3}, {4,0,7}
  , {1,5,6}, {5,1,2} };

  // calculating face normals
  Vec3f v[36];
  Vec3f n[36];
  Face3 f[12];
  Vec3f v0, v1, v2, d0, d1, fn;
  for ( int kf=0, kv=0; kf<12; kf++, kv+=3 )
  {
    v0 = vertices[faces[kf][0]];
    v1 = vertices[faces[kf][1]];
    v2 = vertices[faces[kf][2]];

    d0 = v1 - v0;
    d1 = v2 - v0;

    fn = d0^d1;
    fn.normalize();

    f[kf][0] = kv;
    f[kf][1] = kv+1;
    f[kf][2] = kv+2;

    v[kv]    = v0;
    v[kv+1]  = v1;
    v[kv+2]  = v2;

    n[kv]    = fn; 
    n[kv+1]  = fn; 
    n[kv+2]  = fn; 
  }

  // Create a VertexAttributeSet with vertices and normals
  VertexAttributeSetHandle * vasHdl = CreateHandle(VertexAttributeSetHandle);
  {
    WritableObject<VertexAttributeSet> vas(vasHdl);
    vas->setVertices( v, 36 );
    vas->setNormals( n, 36 );
  }

  // Create a PrimitiveSet, in this case triangles.
  TrianglesHandle * shapeHdl = CreateHandle(TrianglesHandle);
  {
    WritableObject<Triangles> triangles(shapeHdl);
    triangles->setVertexAttributeSet( vasHdl );
    triangles->setFaces(f, 12);
  }

  // Create a material.
  MaterialHandle * materialHdl = CreateHandle(MaterialHandle);
  {
    WritableObject<Material> material(materialHdl);
    material->setAmbientColor( Vec3f(0.f, 0.f, 1.f) );
    material->setDiffuseColor( Vec3f(0.4f, 0.4f, 0.4f) );
    material->setEmissiveColor( Vec3f(0.f, 0.f, 0.f) );
    material->setSpecularColor( Vec3f(1.f, 1.f, 1.f) );
    material->setSpecularExponent( 10.f );
    material->setOpacity( 1.0f );
  }

  StateSetHandle * stateSetHdl = CreateHandle(StateSetHandle);
  WritableObject<StateSet>(stateSetHdl)->addAttribute(materialHdl);

  // Setup the whole scene graph
  GeoNodeHandle * geoNodeHdl = CreateHandle(GeoNodeHandle);
  WritableObject<GeoNode>(geoNodeHdl)->addGeometry(shapeHdl,stateSetHdl);

  GeoNodeHandle * geoNodeHdl2 = geoNodeHdl->clone();

  // Note: 
  // AddChild() will increment the reference count 
  // of the added node.
  Trafo shapeTrafo2;
  shapeTrafo2.setOrientation( Quatf(Vec3f(0.0f, 1.0f, 0.0f ), float(PI_QUARTER)) );
  TransformHandle * pShapeTrafo2 = CreateHandle(TransformHandle);
  {
    WritableObject<Transform> wot(pShapeTrafo2);
    wot->setTrafo( shapeTrafo2 );
    wot->addChild( geoNodeHdl );
  }

  Trafo shapeTrafo;
  shapeTrafo.setOrientation( Quatf(Vec3f(1.0f, 0.0f, 0.0f ), float(PI_QUARTER)) );
  TransformHandle * pShapeTrafo = CreateHandle(TransformHandle);
  {
    WritableObject<Transform> wot(pShapeTrafo);
    wot->setTrafo(shapeTrafo);
    wot->addChild( pShapeTrafo2 );
  }

  Trafo shape2Trafo2;
  shape2Trafo2.setOrientation( Quatf(Vec3f(0.0f, 1.0f, 0.0f ), float(PI_QUARTER)) );
  TransformHandle * pShape2Trafo2 = CreateHandle(TransformHandle);
  {
    WritableObject<Transform> wot(pShape2Trafo2);
    wot->setTrafo(shape2Trafo2);
    wot->addChild(geoNodeHdl2);
  }

  Trafo shape2Trafo;
  shape2Trafo.setOrientation( Quatf(Vec3f(1.0f, 0.0f, 0.0f ), 0.0f) );
  shape2Trafo.setTranslation( Vec3f( -1.0f, 0.0f, 0.0f ) ); 
  TransformHandle * pShape2Trafo = CreateHandle(TransformHandle);
  {
    WritableObject<Transform> wot(pShape2Trafo);
    wot->setTrafo(shape2Trafo);
    wot->addChild(pShape2Trafo2);
  }

  Trafo shape3Trafo;
  shape3Trafo.setOrientation( Quatf(Vec3f(1.0f, 0.0f, 0.0f ), 0.0f) );
  shape3Trafo.setTranslation( Vec3f( 1.0f, 0.0f, 0.0f ) ); 
  TransformHandle * pShape3Trafo = CreateHandle(TransformHandle);
  {
    WritableObject<Transform> wot(pShape3Trafo);
    wot->setTrafo(shape3Trafo);
    wot->addChild( geoNodeHdl2 );
  }

  GroupHandle * rootHdl = CreateHandle(GroupHandle);
  {
    WritableObject<Group> root(rootHdl);
    root->addChild( pShapeTrafo );
    root->addChild( pShape2Trafo );
    root->addChild( pShape3Trafo );
  }

  return( rootHdl );

#else

  // This code is for testing LODs only
  // -----------------------------------

  // setup vertices and faces  
  Vec3f vertices[8] = { Vec3f(-0.2f, -0.2f,  0.2f)
    , Vec3f(-0.2f,  0.2f,  0.2f)
    , Vec3f( 0.2f,  0.2f,  0.2f)
    , Vec3f( 0.2f, -0.2f,  0.2f)
    , Vec3f( 0.2f, -0.2f, -0.2f)
    , Vec3f( 0.2f,  0.2f, -0.2f)
    , Vec3f(-0.2f,  0.2f, -0.2f)
    , Vec3f(-0.2f, -0.2f, -0.2f) };

  Face3 faces[12]     = { {0,2,1}, {2,0,3}
  , {3,5,2}, {5,3,4}
  , {4,6,5}, {6,4,7}
  , {7,1,6}, {1,7,0}
  , {0,4,3}, {4,0,7}
  , {1,5,6}, {5,1,2} };

  // calculating face normals
  Vec3f v[36];
  Vec3f n[36];
  Face3 f[12];
  Vec3f v0, v1, v2, d0, d1, fn;
  for ( int kf=0, kv=0; kf<12; kf++, kv+=3 )
  {
    v0 = vertices[faces[kf][0]];
    v1 = vertices[faces[kf][1]];
    v2 = vertices[faces[kf][2]];

    d0 = v1 - v0;
    d1 = v2 - v0;

    fn = d0^d1;
    fn.normalize();

    f[kf][0] = kv;
    f[kf][1] = kv+1;
    f[kf][2] = kv+2;

    v[kv]    = v0;
    v[kv+1]  = v1;
    v[kv+2]  = v2;

    n[kv]    = fn; 
    n[kv+1]  = fn; 
    n[kv+2]  = fn; 
  }

  // Create a VertexAttributeSet
  VertexAttributeSetHandle * vasHdl = CreateHandle(VertexAttributeSetHandle);
  {
    WritableObject<VertexAttributeSet> vas(vasHdl);
    vas->setVertices(v,36);
    vas->setNormals(n,36);
  }

  // Create a GeoSet, in this case triangles.
  TrianglesHandle * triHdl = CreateHandle(TrianglesHandle);
  {
    WritableObject<Triangles> tri(triHdl);
    tri->setVertexAttributeSet(vasHdl);
    tri->setFaces(f, 12);
  }

  // Create a material.
  MaterialHandle * matHdl = CreateHandle(MaterialHandle);
  {
    WritableObject<Material> mat(matHdl);
    mat->setAmbientColor( Vec3f(1.f, 0.f, 0.f) );
    mat->setDiffuseColor( Vec3f(0.4f, 0.4f, 0.4f) );
    mat->setEmissiveColor( Vec3f(0.f, 0.f, 0.f) );
    mat->setSpecularColor( Vec3f(1.f, 1.f, 1.f) );
    mat->setSpecularExponent( 10.f );
    mat->setOpacity( 1.0f );
  }

  // Create a face attribute. The face attribute is a state attribute 
  // that determines how the following faces will be rendered.
  FaceAttributeHandle * faHdl = CreateHandle(FaceAttributeHandle);

  // Create state set. Add material and render mode.
  StateSetHandle * ssHdl = CreateHandle(StateSetHandle);
  {
    WritableObject<StateSet> ss(ssHdl);
    ss->addAttribute( matHdl );
    ss->addAttribute( faHdl );
  }

  // Setup the whole scene graph
  GeoNodeHandle * levelOneHdl = CreateHandle(GeoNodeHandle);
  WritableObject<GeoNode>(levelOneHdl)->addGeometry( triHdl, ssHdl );

#ifndef LOD_EMPTYGROUPTEST
  GeoNodeHandle * levelTwoHdl = levelOneHdl->clone();
  {
    ReadableObject<StateSet> ss(ReadableObject<GeoNode>(levelTwoHdl)->getStateSet((size_t)0));
    matHdl = dynamic_cast<MaterialHandle*>(ss->getAttributeByObjectCode(OC_MATERIAL));
    WritableObject<Material>(matHdl)->setAmbientColor( Vec3f(0.0f,1.0f,0.0f));
  }
#endif

  GeoNodeHandle * levelThreeHdl = levelOneHdl->clone();
  {
    ReadableObject<StateSet> ss(ReadableObject<GeoNode>(levelTwoHdl)->getStateSet((size_t)0));
    matHdl = dynamic_cast<MaterialHandle*>(ss->getAttributeByObjectCode(OC_MATERIAL));
    WritableObject<Material>(matHdl)->setAmbientColor( Vec3f(0.0f,0.0f,1.0f));
  }

  float ranges[2];
  ranges[0] = .7f;
  ranges[1] = .9f;
  LODHandle * lodHdl = CreateHandle(LODHandle);
  {
    WritableObject<LOD> lod(lodHdl);
    lod->setRanges(&ranges[0],2); 
    lod->addChild( levelOneHdl );   //r
#ifdef LOD_EMPTYGROUPTEST
    lod->addChild( CreateHandle(GroupHandle) );
#else
    lod->addChild( levelTwoHdl );   //g
#endif
    lod->addChild( levelThreeHdl ); //b
  }

  // LOD and stages created
  // now we need a transform
  Trafo trafo;
  trafo.setTranslation( Vec3f(.0f, 1.0f, 1.0f) );
  TransformHandle * transformHdl = CreateHandle(TransformHandle);
  {
    WritableObject<Transform> transform(transformHdl);
    transform->setTrafo(trafo);
    transform->addChild(lodHdl);
  }

  // create root group
  GroupHandle * rootHdl = CreateHandle(GroupHandle);
  {
    WritableObject<Group> root(rootHdl);
    root->addChild(lodHdl);
    root->addChild(transformHdl);
  }

  return( rootHdl );
#endif
}
// SceniX ---------------------- end


// CSimpleViewerDoc serialization

void CSimpleViewerDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// TODO: add storing code here
	}
	else
	{
		// TODO: add loading code here
	}
}


// CSimpleViewerDoc diagnostics

#ifdef _DEBUG
void CSimpleViewerDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CSimpleViewerDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG


void CSimpleViewerDoc::DeleteContents()
{
  // SceniX ---------------------- begin
  // Clean up the scene. 
  if ( m_sceneHdl )
  {
    m_sceneHdl->removeRef();
    m_sceneHdl = NULL;
  }

  // SceniX ---------------------- end
  
  CDocument::DeleteContents();
}
