// SimpleViewerView.cpp : implementation of the CSimpleViewerView class
//

#include "stdafx.h"
#include "SimpleViewer.h"

#include "SimpleViewerDoc.h"
#include "SimpleViewerView.h"

// SceniX ---------------------- begin
#include "nvsg/ErrorHandling.h"
#include "nvsg/PerspectiveCamera.h"
#include "nvsg/Scene.h"
#include "nvsg/DirectedLight.h"

#include "ramfc/GLMFCRAWnd.h"
#include "nvui/TrackballCameraManipulator.h"

#include "nvutil/IOLayer.h"

#include "SilverLiningManager.h"

using namespace nvmath;
using namespace nvsg;
using namespace nvui;
using namespace nvutil;

// enable memory leak detection
#include "nvutil/DbgNew.h"

// SceniX ---------------------- end


// CSimpleViewerView

IMPLEMENT_DYNCREATE(CSimpleViewerView, CView)

BEGIN_MESSAGE_MAP(CSimpleViewerView, CView)
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
  ON_WM_CREATE()
  ON_WM_DESTROY()
  ON_WM_ERASEBKGND()
  ON_WM_SIZE()
  ON_WM_KEYDOWN()
  ON_WM_KEYUP()
  ON_WM_LBUTTONDOWN()
  ON_WM_LBUTTONUP()
  ON_WM_MBUTTONDOWN()
  ON_WM_MBUTTONUP()
  ON_WM_MOUSEMOVE()
  ON_WM_RBUTTONDOWN()
  ON_WM_KILLFOCUS()
END_MESSAGE_MAP()

// CSimpleViewerView construction/destruction

CSimpleViewerView::CSimpleViewerView()
: m_pRA(NULL)
, m_pCamManip(NULL)
, m_pSilverLiningManager(NULL)
{
	// TODO: add construction code here

}

CSimpleViewerView::~CSimpleViewerView()
{
}

BOOL CSimpleViewerView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

// CSimpleViewerView drawing

void CSimpleViewerView::OnDraw(CDC* /*pDC*/)
{
	CSimpleViewerDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
	if (!pDoc)
		return;

  // SceniX ---------------------- begin
  
  // Simply render the scene and react on exceptions.
  try
  {
    m_pRA->renderScene();
  }
  catch(const NVSGException& e) // catch unexpected SceniX errors
  {
    nvutil::IO::errorMessage("SceniX Critical Error", e.getErrorMessage(), false);
    exit(1); // program termination on critical error
  }
  catch(...)
  {
    nvutil::IO::errorMessage("Unknown Error", "An unknown error occurred! Cannot proceed", false);
    exit(1); // program termination on critical error
  }
  // SceniX ---------------------- end
}


// CSimpleViewerView printing

BOOL CSimpleViewerView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CSimpleViewerView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CSimpleViewerView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}


// CSimpleViewerView diagnostics

#ifdef _DEBUG
void CSimpleViewerView::AssertValid() const
{
	CView::AssertValid();
}

void CSimpleViewerView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CSimpleViewerDoc* CSimpleViewerView::GetDocument() const // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CSimpleViewerDoc)));
	return (CSimpleViewerDoc*)m_pDocument;
}
#endif //_DEBUG


// CSimpleViewerView message handlers

int CSimpleViewerView::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
  // SceniX ---------------------- begin
  // Setup RenderArea on create to 
  // detect unsuitable hardware.
  setupRenderArea();
  if ( !m_pRA )
  {
    // in OnCreate we have a chance to reject
    return -1;
  }

  return CView::OnCreate(lpCreateStruct);
  // SceniX ---------------------- end
}

// SceniX ---------------------- begin
void CSimpleViewerView::cleanupRenderArea()
{
  // Clean up silverlining.
  if (m_pSilverLiningManager)
  {
      delete m_pSilverLiningManager;
      m_pSilverLiningManager = NULL;
  }

  // Render area clean up.
  // Important: destroy RA first and raWnd second so the RA can free DCs
  if ( m_pRA ) 
  {
    m_pRA->destroy();
    delete m_pRA;
    m_pRA = NULL;
  }

}
// SceniX ---------------------- end

// SceniX ---------------------- begin
void CSimpleViewerView::setupRenderArea()
{
  // Throw away existing render area stuff.
  cleanupRenderArea();

  // Create a new OpenGL render area.
  GLMFCRAWnd* pRA;
  VERIFY(pRA = new GLMFCRAWnd);  
  
  if (pRA)
  {
    RECT rect = {0,0,0,0};
    if (pRA->Create("STATIC", NULL, WS_CHILD, rect, this, (UINT)RenderArea::createNewRenderAreaID()))
    {
      // downcast, since we only need base class interface from now on
      m_pRA=pRA;

      if ( m_pRA->init() )
      {
        // Adjust size of render area.
        RECT rect;
        GetClientRect(&rect);
        m_pRA->setViewportSize( rect.right, rect.bottom );
      }
      else
      {
        AfxMessageBox("Can't setup render area!", MB_OK | MB_ICONEXCLAMATION );
      
        delete m_pRA;
        m_pRA = NULL;
        
        exit(1);
      }
    }
    else
    {
      AfxMessageBox("Can't create render window!", MB_OK | MB_ICONEXCLAMATION );
      exit(1);
    }
  }

  m_pSilverLiningManager = new SilverLiningManager;
}

void CSimpleViewerView::setupAtmosphericConditions(SilverLining::Atmosphere *atm)
{
    if (!atm) return;

    // Set the visibility so clouds fade out in the distance
    atm->GetConditions()->SetVisibility(30000.0);

    // Set the simulated location to somewhere near Seattle
    SilverLining::Location loc;
    loc.SetAltitude(0);
    loc.SetLatitude(47);
    loc.SetLongitude(-122);
    atm->GetConditions()->SetLocation(loc);

    // Set the time to midsummer - it's important that the time zone
    // specified matches the position above!
    SilverLining::LocalTime tm;
    tm.SetTimeZone(PST);
    tm.SetYear(2010);
    tm.SetMonth(7);
    tm.SetDay(10);
    tm.SetHour(17);
    tm.SetMinutes(0);
    tm.SetSeconds(0);
    atm->GetConditions()->SetTime(tm);

    // Add some cirrus clouds
    SilverLining::CloudLayer *cirrusCloudLayer = SilverLining::CloudLayerFactory::Create(CIRRUS_FIBRATUS);
    cirrusCloudLayer->SetIsInfinite(true);
    cirrusCloudLayer->SetBaseAltitude(8000);
    cirrusCloudLayer->SetThickness(0);
    cirrusCloudLayer->SetBaseLength(100000);
    cirrusCloudLayer->SetBaseWidth(100000);
    cirrusCloudLayer->SetLayerPosition(0, 0);
    cirrusCloudLayer->SeedClouds(*atm);

    atm->GetConditions()->AddCloudLayer(cirrusCloudLayer);

    // And some cumulus congestus
    SilverLining::CloudLayer *cumulusCongestusLayer;
    cumulusCongestusLayer = SilverLining::CloudLayerFactory::Create(CUMULUS_CONGESTUS);
    cumulusCongestusLayer->SetIsInfinite(true);
    cumulusCongestusLayer->SetBaseAltitude(1500);
    cumulusCongestusLayer->SetThickness(100);
    cumulusCongestusLayer->SetBaseLength(30000);
    cumulusCongestusLayer->SetBaseWidth(30000);
    cumulusCongestusLayer->SetDensity(0.4);
    cumulusCongestusLayer->SetLayerPosition(0, 0);
    cumulusCongestusLayer->SeedClouds(*atm);

    atm->GetConditions()->AddCloudLayer(cumulusCongestusLayer);

}

// SceniX ---------------------- end

void CSimpleViewerView::OnDestroy()
{
  // SceniX ---------------------- begin
  cleanupRenderArea();

  // Clean up camera manipulator.
  if (m_pCamManip)
  {
    delete (m_pCamManip);
    m_pCamManip = NULL;
  }
  // SceniX ---------------------- end

  CView::OnDestroy();
}

void CSimpleViewerView::OnInitialUpdate()
{
  // SceniX ---------------------- begin
  // Setup RenderArea on create to 
  // detect unsuitable hardware.
  
  // NOTE: As this is no loader application, we don't need to re-setup a new
  // render area here. In more general cases, where we need to setup a new render 
  // area with each OnInitialUpdate call, the application needs to take care of
  // OpenGL context sharing, if desired.

  // setupRenderArea();
  
  if ( m_pRA ) 
  {
    if ( ! m_pCamManip )
    {
      VERIFY( m_pCamManip = new TrackballCameraManipulator );
    }
    m_pCamManip->setRenderArea(m_pRA);

    SceneHandle * sceneHdl = GetDocument()->getScene();
    if ( sceneHdl )
    {
      // Create camera.
      CameraHandle * cameraHdl = CreateHandle(PerspectiveCameraHandle);
      {
        WritableObject<Camera> camera(cameraHdl);

        // Make the (world space) window aspect ratio 
        // correspond to the (pixel space) viewport aspect ratio
        camera->setName( "ViewCamera" );
        camera->setAspectRatio( m_pRA->getAspectRatio() );
        
        // Position the camera.
        camera->setPosition(nvmath::Vec3f(0.0f, 0.0f, 7.0f));

        camera->setAutoClipPlanes(true);
      }

      ViewStateHandle * viewStateHdl = CreateHandle(ViewStateHandle);
      WritableObject<ViewState>(viewStateHdl)->setCamera( cameraHdl );

      m_pRA->setSceneData( sceneHdl, viewStateHdl );

      if (m_pSilverLiningManager)
      {
         NodeHandle *root = ReadableObject<Scene>(sceneHdl)->getRootNode();
         GroupHandle *groupHdl = dynamic_cast<GroupHandle *>(root);
         if (groupHdl)
         {
             // Add a directional light source (managed via SilverLiningManager each frame)
             DirectedLightHandle *lightHdl = CreateHandle(DirectedLightHandle);
             {
                WritableObject<DirectedLight> light(lightHdl);
                light->setAmbientColor(Vec3f(1.0f, 1.0f, 1.0f));
                light->setDiffuseColor(Vec3f(1.0f, 1.0f, 1.0f));
                light->setSpecularColor(Vec3f(0.0f, 0.0f, 0.0f));
             }
             WritableObject<Group>(groupHdl)->addLightSource(lightHdl);
             SilverLining::Atmosphere *atm = m_pSilverLiningManager->AttachToRoot(groupHdl, cameraHdl, lightHdl);
             setupAtmosphericConditions(atm);
         }
      }
    }
  }
  // SceniX ---------------------- end

  // call base class
  CView::OnInitialUpdate();
}

BOOL CSimpleViewerView::OnEraseBkgnd(CDC* pDC)
{
  // SceniX ---------------------- begin

  // We are parent of an OGL window, so do nothing in here. 
  // The render Area will handle this.
  return TRUE;
  
  // SceniX ---------------------- end
}

void CSimpleViewerView::OnSize(UINT nType, int cx, int cy)
{
  CView::OnSize(nType, cx, cy);

  // SceniX ---------------------- begin
  // broadcast to render area if any
  if ( m_pRA )
  {
    m_pRA->setViewportSize( cx, cy );
  }
  // SceniX ---------------------- end

}

void CSimpleViewerView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  // SceniX ---------------------- begin
  // we do not want to catch key repeats
  if (!(nFlags & 0x4000) && m_pCamManip)
  {
    switch(nChar) 
    {
    case VK_SHIFT:
      m_pCamManip->keyDown(NVSG_SHIFT);
      break;
    case VK_CONTROL:
      m_pCamManip->keyDown(NVSG_CONTROL);
      break;
    }
  }
  // SceniX ---------------------- end

  CView::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CSimpleViewerView::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags)
{
  // SceniX ---------------------- begin
  if (m_pCamManip)
  {
    switch(nChar) 
    {
    case VK_SHIFT:
      m_pCamManip->keyUp(NVSG_SHIFT);
      break;
    case VK_CONTROL:
      m_pCamManip->keyUp(NVSG_CONTROL);
      break;
    }
  }
  // SceniX ---------------------- end

  CView::OnKeyUp(nChar, nRepCnt, nFlags);
}

void CSimpleViewerView::OnLButtonDown(UINT nFlags, CPoint point)
{
  // SceniX ---------------------- begin
  if (m_pCamManip)
  {
    m_pCamManip->mouseButtonDown(NVSG_LEFT, point.x, point.y);
  }
  // SceniX ---------------------- end

  CView::OnLButtonDown(nFlags, point);
}

void CSimpleViewerView::OnLButtonUp(UINT nFlags, CPoint point)
{
  // SceniX ---------------------- begin
  if (m_pCamManip)
  {
    m_pCamManip->mouseButtonUp(NVSG_LEFT, point.x, point.y);
  }
  // SceniX ---------------------- end

  CView::OnLButtonUp(nFlags, point);
}

void CSimpleViewerView::OnMButtonDown(UINT nFlags, CPoint point)
{
  // SceniX ---------------------- begin
  if (m_pCamManip)
  {
    m_pCamManip->mouseButtonDown(NVSG_MIDDLE, point.x, point.y);
  }
  // SceniX ---------------------- end

  CView::OnMButtonDown(nFlags, point);
}

void CSimpleViewerView::OnMButtonUp(UINT nFlags, CPoint point)
{
  // SceniX ---------------------- begin
  if (m_pCamManip)
  {
    m_pCamManip->mouseButtonUp(NVSG_MIDDLE, point.x, point.y);
  }
  // SceniX ---------------------- end

  CView::OnMButtonUp(nFlags, point);
}

void CSimpleViewerView::OnMouseMove(UINT nFlags, CPoint point)
{
  // SceniX ---------------------- begin
  if (m_pCamManip)
  {
    m_pCamManip->mouseMotion(point.x, point.y);
  }
  // SceniX ---------------------- end

  CView::OnMouseMove(nFlags, point);
}

void CSimpleViewerView::OnRButtonDown(UINT nFlags, CPoint point)
{
  // SceniX ---------------------- begin

  if ( m_pRA )
  {
    // Reset view (view all)
    SceneHandle * sceneHdl = GetDocument()->getScene();
    if (sceneHdl && ReadableObject<ViewState>(m_pRA->getViewState())->getCamera())
    {
      // save undo data
      WritableObject<Camera> camera(ReadableObject<ViewState>(m_pRA->getViewState())->getCamera() );
      NodeHandle * rootHdl = ReadableObject<Scene>(sceneHdl)->getRootNode();
      if (rootHdl && camera)
      {
          camera->setPosition(nvmath::Vec3f(0.0, 0.0, 7.0f));
          camera->setOrientation(nvmath::Vec3f(0.0f, 0.0f, -1.0f), nvmath::Vec3f(0.0f, 1.0f, 0.0f));
      }
    }

    // Force redraw.
    // Note:
    // Do not call the drawing routines of the render area directly
    // except you are in OnDraw().
    m_pRA->triggerRedraw();
  }

  // SceniX ---------------------- end

  CView::OnRButtonDown(nFlags, point);
}

void CSimpleViewerView::OnKillFocus(CWnd* pNewWnd)
{
  CView::OnKillFocus(pNewWnd);

  // SceniX ---------------------- begin

  // Since MFC tells key up events etc only to the 
  // window that has the focus we need to reset the 
  // key and button state here. 
  if (m_pCamManip)
  {
    m_pCamManip->reset();
  }

  // SceniX ---------------------- end
}

// SceniX ---------------------- begin
BOOL CSimpleViewerView::doIdleHandling()
{
  if (m_pRA && m_pCamManip)
  {
    if (m_pCamManip->updateFrame())
    {
      m_pRA->triggerRedraw();
    }
  }

  return FALSE; // no more to do
}
// SceniX ---------------------- end
